﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace bankchurn
{
    class KARNO
    {
        private List<client> _clients = new List<client>();
        public List<double> _targets = new List<double>();
        private List<UBranch> _ulist = new List<UBranch>();
        private U _bigU = new U();
        private double _maxtarget;
        private double _mintarget;
        public List<double> _model = new List<double>();
        private List<bool> _isTraining = new List<bool>();
        private double _mu = 0.1;

        public KARNO(List<client> client, List<double> targets)
        {
            _clients = client;
            _targets = targets;
            _mintarget = _targets.Min();
            _maxtarget = _targets.Max();
            for (int i = 0; i < _targets.Count; ++i)
            {
                _model.Add(Double.MaxValue); //fill blank
            }
        }

        public void FillModel()
        {
            for (int i = 0; i < _model.Count; ++i)
            {
                if (false == _isTraining[i])
                {
                    client C = _clients[i];
                    _model[i] = PredictElement(C);
                }
            }
        }

        public void GetScore()
        {
            int right = 0;
            int cnt = 0;
            for (int i = 0; i < _model.Count; ++i)
            {
                if (_model[i] >= Double.MaxValue - 1)
                {
                    Console.WriteLine("Warning: model values may not be assigned correctly!");
                    Environment.Exit(0);
                }
                double m = _model[i];
                double t = _targets[i];
                ++cnt;
                if (m > 0.0 && t > 0.0) ++right;
                if (m <= 0.0 && t <= 0.0) ++right;
            }
            Console.WriteLine("Right predictions = {0}, from {1} for unseen data.", right, cnt);
        }

        public double GetRMSE()
        {
            double error = 0.0;
            int cnt = 0;
            for (int i = 0; i < _model.Count; ++i)
            {
                double m = _model[i];
                double t = _targets[i];
                error += (m - t) * (m - t);
                ++cnt;
            }
            error /= cnt;
            error = Math.Sqrt(error);
            return error;
        }

        public double PredictElement(client C)
        {
            double[] v = GetVector(C);
            double predictedbyKAelement = _bigU.GetU(v);
            return predictedbyKAelement;
        }

        private double[] GetVector(client C)
        {
            int size = _ulist.Count;
            double[] vector = new double[size];
            for (int i = 0; i < size; ++i)
            {
                vector[i] = _ulist[i].GetU(C);
            }
            return vector;
        }

        public void ExecuteUpdateSteps(int nSteps)
        {
            int N = _clients.Count;
            double error = (_maxtarget - _mintarget) / 150.0;
            double training_error = 0.0;

            for (int step = 0; step < nSteps; ++step)
            {
                double accuracy = 0.0;
                int cnt = 0;
                for (int i = 0; i < N; ++i)
                {
                    if (false == _isTraining[i]) continue;
                    client C = _clients[i];
                    double[] v = GetVector(C);
                    double diff = _targets[i] - _bigU.GetU(v);

                    if (Math.Abs(diff) > error)
                    {
                        double diffPerLayer = diff / _bigU.GetNumberOfLayers();
                        for (int k = 0; k < _ulist.Count; ++k)
                        {
                            double derrivative = _bigU.GetDerrivative(k, v[k]);
                            if (Math.Abs(derrivative) < 0.1)
                            {
                                if (derrivative < 0.0) derrivative = -0.1;
                                else derrivative = 0.1;
                            }
                            _ulist[k].Update(C, diffPerLayer / derrivative, _mu);
                        }
                        v = GetVector(C);
                        diff = _targets[i] - _bigU.GetU(v);
                        _bigU.Update(diff, v, _mu);
                    }
                    accuracy += diff * diff;
                    ++cnt;
                }

                accuracy /= cnt;
                accuracy = Math.Sqrt(accuracy);
                accuracy /= (_maxtarget - _mintarget);
                //if (0 == step % 10)
                //{
                //    Console.WriteLine("Error in big operator {0:0.0000}, step {1}", accuracy, step);
                //}
                training_error = accuracy;
            }
            //Console.WriteLine("Training is completed, relative error {0:0.0000}", training_error);
        }

        public void MakeTrainTestSplit(int block, int blocks)
        {
            _isTraining.Clear();
            int size = _targets.Count;
            int blocksize = size / blocks;
            for (int i = 0; i < size; ++i)
            {
                if (i >= block * blocksize && i < (block + 1) * blocksize)
                {
                    _isTraining.Add(false);
                }
                else
                {
                    _isTraining.Add(true);
                }
            }
        }

        public void GenerateInitialOperators(int M, Random rnd)
        {
            _ulist.Clear();
            for (int counter = 0; counter < M; ++counter)
            {
                UBranch u = new UBranch();
                u.Initialize();
                for (int count = 1; count < 4; ++count)
                {
                    double accuracy = 0.0;
                    int cnt = 0;
                    for (int i = 0; i < _clients.Count; ++i)
                    {
                        if (false == _isTraining[i]) continue;
                        double v = 1.0;
                        if (rnd.Next() % 100 > 50.0) v = -1.0;
                        client C = _clients[i];
                        double diff = v - u.GetU(C);
                        u.Update(C, diff, _mu);
                        accuracy += diff * diff;
                        ++cnt;
                    }
                    accuracy /= cnt;
                    accuracy = Math.Sqrt(accuracy);
                    //Console.WriteLine("Error for initial operator {0}, {1:0.0000}", counter, accuracy);
                }
                _ulist.Add(u);
            }
            //Console.WriteLine("Initialization of root operator");
            double[] minbig = new double[M];
            double[] maxbig = new double[M];
            for (int k = 0; k < M; ++k)
            {
                minbig[k] = _mintarget;
                maxbig[k] = _maxtarget;
            }
            _bigU.Clear();
            _bigU.Initialize(M, 4, minbig, maxbig);
            for (int step = 0; step < 2; ++step)
            {
                double accuracy = 0.0;
                int cnt = 0;
                for (int i = 0; i < _clients.Count; ++i)
                {
                    if (false == _isTraining[i]) continue;
                    client C = _clients[i];
                    double[] v = GetVector(C);
                    double diff = _targets[i] - _bigU.GetU(v);
                    _bigU.Update(diff, v, _mu);
                    accuracy += diff * diff;
                    ++cnt;
                }
                accuracy /= cnt;
                accuracy = Math.Sqrt(accuracy);
                accuracy /= (_maxtarget - _mintarget);
                //Console.WriteLine("Error in initial approximation of root operator {0:0.0000}, step {1}", accuracy, step);
            }
        }
    }
}
